/*#################################################################################################
##  |\   ======   /|  ##     Copyright (C) 2013-2014  Maurizio Sartori     ##  |\   ======   /|  ##
##  |__\========/__|  ##  ===============================================  ##  |__\========/__|  ##
##  |\            /|  ##     Written by:  Maurizio Sartori                 ##  |\            /|  ##
##  |  \        /  |  ##     e-mail:      masar@MasarLabs.com              ##  |  \        /  |  ##
##  |    \    /    |  ##     Site:        http://www.masarlabs.com         ##  |    \    /    |  ##
##  |    /=\/=\    |  ##  ===============================================  ##  |    /=\/=\    |  ##
##  |  /========\  |  ##     License GPLv3+: GNU GPL version 3 or later    ##  |  /========\  |  ##
##  |/   ======   \|  ##          http://gnu.org/licenses/gpl.html         ##  |/   ======   \|  ##
###################################################################################################
## $Id: pat.c 4122 2014-02-12 12:10:23Z masar@masar $
#################################################################################################*/

static char Psz_SvnId[] __attribute__ ((unused)) =
    "$Id: pat.c 4122 2014-02-12 12:10:23Z masar@masar $";

#include "config.h"
#include "network.h"
#include "patd.h"
#include "util.h"

#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//************************************************************************************************/
//* Data
//************************************************************************************************/
static long i_Count;
static long i_Fragment;
static long i_MaxTTL;
static long i_Return;
static long i_Size;
static long i_TimeOut;

#define TYPE_ANY  0
#define TYPE_IP4  4
#define TYPE_IP6  6
static int i_Type;

uint8_t aui8_Buffer[IP_MAXPACKET + 1];
long i_Send;
long i_StartTTL;

/*************************************************************************************************/
/** Exit on error.
**************************************************************************************************/
static void chk (int iRc, const char * str)
   {
   if (iRc >= 0) {
      return;
      }
   fprintf(stderr, "%s: %s\n", str, strerror(errno));
   exit(EXIT_FAILURE);
   }

/*************************************************************************************************/
/** Processes.
**************************************************************************************************/
static void process (int fdServer, uint8_t * pAddr, int iCommand)
   {
   int iAddrLen;
   int iReqSize;
   int iResSize;

   if (iCommand == COMMAND_IP4) {
      fprintf(stdout, "pat4 %i.%i.%i.%i\n",
                      pAddr[ 0], pAddr[ 1], pAddr[ 2], pAddr[ 3]);
      iAddrLen = 4;
      iReqSize = COMMAND_IP4_SIZE;
      iResSize = RESPONSE_IP4_SIZE;
      }
   else {
      fprintf(stdout, "pat6 %02X%02X:%02X%02X:%02X%02X:%02X%02X:"
                      "%02X%02X:%02X%02X:%02X%02X:%02X%02X\n",
                      pAddr[ 0], pAddr[ 1], pAddr[ 2], pAddr[ 3],
                      pAddr[ 4], pAddr[ 5], pAddr[ 6], pAddr[ 7],
                      pAddr[ 8], pAddr[ 8], pAddr[10], pAddr[11],
                      pAddr[12], pAddr[13], pAddr[14], pAddr[15]);
      iAddrLen = 16;
      iReqSize = COMMAND_IP6_SIZE;
      iResSize = RESPONSE_IP6_SIZE;
      }

   int iCount = 0;

   aui8_Buffer[iCount++] = iCommand;                        // Command
   aui8_Buffer[iCount++] = (i_Send      >>  0) & 0xFF;      // Send (max 63) default 3
   aui8_Buffer[iCount++] = (i_Return    >>  0) & 0xFF;      // Return (max 63) default 1
   aui8_Buffer[iCount++] = (i_StartTTL  >>  0) & 0xFF;      // TTL (max 63) default 63
   aui8_Buffer[iCount++] = (i_MaxTTL    >>  0) & 0xFF;      // TTL (max 63) default 63
   aui8_Buffer[iCount++] = (i_Size      >>  8) & 0xFF;      // size (max 0x7FF0) default 0
   aui8_Buffer[iCount++] = (i_Size      >>  0) & 0xFF;
   aui8_Buffer[iCount++] = i_Fragment ? 1 : 0;              // Fragment
   aui8_Buffer[iCount++] = (i_TimeOut   >>  8) & 0xFF;      // TimeOut ms (max 2000) default 1000
   aui8_Buffer[iCount++] = (i_TimeOut   >>  0) & 0xFF;

   for (int j = 0; j < iAddrLen; j += 1) {
      aui8_Buffer[iCount++] = pAddr[j];
      }

   if (iCount != iReqSize) {
      chk(-1, "Invalid Message Size");
      }

   chk(writeAll(fdServer, aui8_Buffer, iCount), "Cannot write to server");

   for (;;) {
      chk(readAll(fdServer, aui8_Buffer, iResSize), "Cannot read from server");

      uint8_t iResult = aui8_Buffer[0];
      uint8_t iSeq    = aui8_Buffer[1];
      uint32_t lTime = (aui8_Buffer[2] << 24) +
                       (aui8_Buffer[3] << 16) +
                       (aui8_Buffer[4] <<  8) +
                       (aui8_Buffer[5] <<  0);

      switch (iResult) {
         case RESULT_OK:
         case RESULT_OK_LAST:
         case RESULT_TRACE:
         case RESULT_TRACE_LAST:
            if (iCommand == COMMAND_IP4) {
               fprintf(stdout, "   %c %i.%i.%i.%i seq=%i time=%lu.%06lu ms\n",
                               iResult,
                               aui8_Buffer[6], aui8_Buffer[7], aui8_Buffer[8], aui8_Buffer[9],
                               iSeq, lTime / 1000000L, lTime % 1000000L);
               }
            else {
               fprintf(stdout, "   %c %02X%02X:%02X%02X:%02X%02X:%02X%02X:"
                               "%02X%02X:%02X%02X:%02X%02X:%02X%02X seq=%i time=%lu.%06lu ms\n",
                               iResult,
                               aui8_Buffer[ 6], aui8_Buffer[ 7], aui8_Buffer[ 8], aui8_Buffer[ 9],
                               aui8_Buffer[10], aui8_Buffer[11], aui8_Buffer[12], aui8_Buffer[13],
                               aui8_Buffer[14], aui8_Buffer[15], aui8_Buffer[16], aui8_Buffer[17],
                               aui8_Buffer[18], aui8_Buffer[19], aui8_Buffer[20], aui8_Buffer[21],
                               iSeq, lTime / 1000000L, lTime % 1000000L);
               }

            if ((iResult == RESULT_OK_LAST) || (iResult == RESULT_TRACE_LAST)) {
               return;
               }
            break;

         case RESULT_ERROR:
         case RESULT_ERROR_LAST:
            if (iCommand == COMMAND_IP4) {
               fprintf(stdout, "   %c *.*.*.* seq=%i timeout!\n", iResult, iSeq);
               }
            else {
               fprintf(stdout, "   %c *:*:*:*:*:*:*:* seq=%i timeout!\n", iResult, iSeq);
               }
            if (iResult == RESULT_ERROR_LAST) {
               return;
               }
            break;

         default:
            fprintf(stderr, "Invalid response\n");
            return;
         }
      }
   }

/*************************************************************************************************/
/** Processes an address.
**************************************************************************************************/
static void processAddress (int fdServer, const char * strAddr)
   {
   struct addrinfo aiHints;
   memset(&aiHints, 0, sizeof(aiHints));

   aiHints.ai_family = AF_UNSPEC;         /* Return IPv4 and IPv6 choices */
   aiHints.ai_socktype = SOCK_STREAM;
   aiHints.ai_protocol = IPPROTO_IP;

   struct addrinfo * paiAddresses = NULL;
   int iRc = getaddrinfo(strAddr, NULL, &aiHints, &paiAddresses);
   if (iRc != 0) {
      fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(iRc));
      return;
      }

   int iProcessed = 0;
   struct addrinfo * pai = paiAddresses;
   while ((pai != NULL) && ((i_Count == 0) || (iProcessed < i_Count))) {
      if (pai->ai_family == AF_INET) {
         uint8_t * pAddr = (uint8_t *) &(((struct sockaddr_in *)pai->ai_addr)->sin_addr.s_addr);
         if ((i_Type == TYPE_ANY) || (i_Type == TYPE_IP4)) {
            process(fdServer, pAddr, COMMAND_IP4);
            iProcessed += 1;
            }
         else {
            fprintf(stdout, "pat4 %i.%i.%i.%i ignored\n",
                            pAddr[ 0], pAddr[ 1], pAddr[ 2], pAddr[ 3]);
            }
         }
      else if (pai->ai_family == AF_INET6) {
         uint8_t * pAddr = ((struct sockaddr_in6 *)pai->ai_addr)->sin6_addr.s6_addr;
         if ((i_Type == TYPE_ANY) || (i_Type == TYPE_IP6)) {
            process(fdServer, pAddr, COMMAND_IP6);
            iProcessed += 1;
            }
         else {
            fprintf(stdout, "pat6 %02X%02X:%02X%02X:%02X%02X:%02X%02X:"
                            "%02X%02X:%02X%02X:%02X%02X:%02X%02X ignored\n",
                            pAddr[ 0], pAddr[ 1], pAddr[ 2], pAddr[ 3],
                            pAddr[ 4], pAddr[ 5], pAddr[ 6], pAddr[ 7],
                            pAddr[ 8], pAddr[ 8], pAddr[10], pAddr[11],
                            pAddr[12], pAddr[13], pAddr[14], pAddr[15]);
            }
         }
      else {
         fprintf(stderr, "Unknown address type %i, %i, %i\n",
                 pai->ai_family, pai->ai_socktype, pai->ai_protocol);
         }

      pai = pai->ai_next;
      }
   freeaddrinfo(paiAddresses);
   }

/*************************************************************************************************/
/** Main entry point.
**************************************************************************************************/
int main (int argc, const char *argv[])
   {
   const char * strHost = "127.0.0.1";
   const char * strPort = "7865";
   i_Send     = 3;            // Ping
   i_StartTTL = 63;

   if (argc == 1) {
      fprintf(stderr, "%s: missing arguments. \"%s -h\" for help.\n", argv[0], argv[0]);
      return EXIT_FAILURE;
      }

   for (int i = 1; i < argc; i += 1) {
      if (!strcmp(argv[i], "--server")) {
         i += 1;
         if (i == argc) {
            fprintf(stderr, "missing server argument\n");
            return EXIT_FAILURE;
            }
         strHost = argv[i];
         }
      else if (!strcmp(argv[i], "--port")) {
         i += 1;
         if (i == argc) {
            fprintf(stderr, "missing port argument\n");
            return EXIT_FAILURE;
            }
         strPort = argv[i];
         }
      else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")) {
         fprintf(stderr,
            "Usage: %s [option]... [host]...\n"
            "Options:\n"
            "        --server HOST     set the \"patd\" server address (default: localhost)\n"
            "        --port PORT       set the \"patd\" server port (default: 7865)\n"
            "   -P,  --ping            do a Ping (default): same as \"--send 3 --ttl 63\"\n"
            "   -T,  --trace           do a Traceroute: same as \"--send 63 --ttl 0\"\n"
            "   -4,  --ipv4            process only IPv4 adresses\n"
            "   -6,  --ipv6            process only IPv6 adresses\n"
            "   -A,  --all             process IPv4 and IPv6 adresses (default)\n"
            "   -h,  --help            display this help and exit\n"
            "   -V,  --version         display version information and exit\n"
            "   -a,  --addresses NUM   set how many address to check for each host (default: 0=all)\n"
            "   -d,  --dont-fragment   don't fragment the packet (default)\n"
            "   -f,  --fragment        allow fragmenting the packet\n"
            "   -l,  --length NUM      set the packet data length (default: 0)\n"
            "   -o,  --timeout NUM     set the timeout for replyes in milliseconds (default: 1000)\n"
            "   -r,  --return NUM      set the ok replyes to return (default: 1)\n"
            "   -s,  --send NUM        set the number of query to send (default: 3)\n"
            "   -t,  --ttl NUM         set the initial TimeToLive value (default: 63)\n",
            argv[0]);
         return EXIT_FAILURE;
         }
      else if (!strcmp(argv[i], "-V") || !strcmp(argv[i], "--version")) {
         fprintf(stderr,
            "%s - MasarLabs PingAndTraceDaemon " VERSION "\n"
            "\n"
            "  Copyright (C) 2013-2014 MasarLabs.com\n"
            "  Written by Maurizio Sartori <masar@MasarLabs.com>\n"
            "\n"
            "  License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.\n"
            "  This is free software: you are free to change and redistribute it.\n"
            "  There is NO WARRANTY, to the extent permitted by law.\n",
            argv[0]);
         return EXIT_FAILURE;
         }
      }

   int fdServer = openTcpClientSocket(strHost, strPort);
   snprintf((char *) aui8_Buffer, sizeof(aui8_Buffer),
            "failed to connect to server %s:%s", strHost, strPort);
   chk(fdServer, (char*) aui8_Buffer);

   char * strEnd;
   for (int i = 1; i < argc; i += 1) {
      if (!strcmp(argv[i], "--server")) {
         i += 1;
         // Yet done
         }
      else if (!strcmp(argv[i], "--port")) {
         i += 1;
         // Yet done
         }
      else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")) {
         // Yet done
         }
      else if (!strcmp(argv[i], "-V") || !strcmp(argv[i], "--version")) {
         // Yet done
         }
      else if (!strcmp(argv[i], "-4") || !strcmp(argv[i], "--ipv4")) {
         i_Type = TYPE_IP4;
         }
      else if (!strcmp(argv[i], "-6") || !strcmp(argv[i], "--ipv6")) {
         i_Type = TYPE_IP6;
         }
      else if (!strcmp(argv[i], "-A") || !strcmp(argv[i], "--all")) {
         i_Type = TYPE_ANY;
         }
      else if (!strcmp(argv[i], "-a") || !strcmp(argv[i], "--addresses")) {
         i += 1;
         if (i == argc) {
            fprintf(stderr, "missing addresses argument: number required\n");
            return EXIT_FAILURE;
            }
         errno = 0;
         i_Count = strtol(argv[i], &strEnd, 10);
         if (errno || *strEnd) {
            fprintf(stderr, "invalid addresses argument: number required\n");
            return EXIT_FAILURE;
            }
         if ((i_Count < 0) || (i_Count > 0x7FFF)) {
            fprintf(stderr, "out of range addresses argument: min=0 max=0x7FFF\n");
            return EXIT_FAILURE;
            }
         }
      else if (!strcmp(argv[i], "-d") || !strcmp(argv[i], "--dont-fragment")) {
         i_Fragment = 0;
         }
      else if (!strcmp(argv[i], "-f") || !strcmp(argv[i], "--fragment")) {
         i_Fragment = 1;
         }
      else if (!strcmp(argv[i], "-l") || !strcmp(argv[i], "--length")) {
         i += 1;
         if (i == argc) {
            fprintf(stderr, "missing length argument: number required\n");
            return EXIT_FAILURE;
            }
         errno = 0;
         i_Size = strtol(argv[i], &strEnd, 10);
         if (errno || *strEnd) {
            fprintf(stderr, "invalid length argument: number required\n");
            return EXIT_FAILURE;
            }
         if ((i_Size < 0) || (i_Size > 0x7FF0)) {
            fprintf(stderr, "out of range length argument: min=0 max=0x7FF0\n");
            return EXIT_FAILURE;
            }
         }
      else if (!strcmp(argv[i], "-o") || !strcmp(argv[i], "--timeout")) {
         i += 1;
         if (i == argc) {
            fprintf(stderr, "missing timeout argument: number required\n");
            return EXIT_FAILURE;
            }
         errno = 0;
         i_TimeOut = strtol(argv[i], &strEnd, 10);
         if (errno || *strEnd) {
            fprintf(stderr, "invalid timeout argument: number required\n");
            return EXIT_FAILURE;
            }
         if ((i_TimeOut <= 0) || (i_TimeOut > 2000)) {
            fprintf(stderr, "out of range timeout argument: min=1 max=2000\n");
            return EXIT_FAILURE;
            }
         }
      else if (!strcmp(argv[i], "-r") || !strcmp(argv[i], "--return")) {
         i += 1;
         if (i == argc) {
            fprintf(stderr, "missing return argument: number required\n");
            return EXIT_FAILURE;
            }
         errno = 0;
         i_Return = strtol(argv[i], &strEnd, 10);
         if (errno || *strEnd) {
            fprintf(stderr, "invalid return argument: number required\n");
            return EXIT_FAILURE;
            }
         if ((i_Return < 0) || (i_Return > 63)) {
            fprintf(stderr, "out of range return argument: min=1 max=63\n");
            return EXIT_FAILURE;
            }
         }
      else if (!strcmp(argv[i], "-s") || !strcmp(argv[i], "--send")) {
         i += 1;
         if (i == argc) {
            fprintf(stderr, "missing send argument: number required\n");
            return EXIT_FAILURE;
            }
         errno = 0;
         i_Send = strtol(argv[i], &strEnd, 10);
         if (errno || *strEnd) {
            fprintf(stderr, "invalid send argument: number required\n");
            return EXIT_FAILURE;
            }
         if ((i_Send < 0) || (i_Send > 63)) {
            fprintf(stderr, "out of range send argument: min=1 max=63\n");
            return EXIT_FAILURE;
            }
         }
      else if (!strcmp(argv[i], "-t") || !strcmp(argv[i], "--ttl")) {
         i += 1;
         if (i == argc) {
            fprintf(stderr, "missing ttl argument: number required\n");
            return EXIT_FAILURE;
            }
         errno = 0;
         i_MaxTTL = strtol(argv[i], &strEnd, 10);
         if (errno || *strEnd) {
            fprintf(stderr, "invalid ttl argument: number required\n");
            return EXIT_FAILURE;
            }
         if ((i_MaxTTL < 0) || (i_MaxTTL > 63)) {
            fprintf(stderr, "out of range ttl argument: min=0 max=63\n");
            return EXIT_FAILURE;
            }
         }
      else if (!strcmp(argv[i], "-T") || !strcmp(argv[i], "--trace")) {
         i_Send     = 63;
         i_StartTTL = 0;
         }
      else if (!strcmp(argv[i], "-P") || !strcmp(argv[i], "--ping")) {
         i_Send     = 3;
         i_StartTTL = 63;
         }
      else if (argv[i][0] == '-') {
         fprintf(stderr, "%s: invalid parameter: %s\n", argv[0], argv[i]);
         return EXIT_FAILURE;
         }
      else {
         processAddress(fdServer, argv[i]);
         }
      }

   close (fdServer);
   return EXIT_SUCCESS;
   }

/*************************************************************************************************/
/*************************************************************************************************/
