/*#################################################################################################
##  |\   ======   /|  ##     Copyright (C) 2013-2014  Maurizio Sartori     ##  |\   ======   /|  ##
##  |__\========/__|  ##  ===============================================  ##  |__\========/__|  ##
##  |\            /|  ##     Written by:  Maurizio Sartori                 ##  |\            /|  ##
##  |  \        /  |  ##     e-mail:      masar@MasarLabs.com              ##  |  \        /  |  ##
##  |    \    /    |  ##     Site:        http://www.masarlabs.com         ##  |    \    /    |  ##
##  |    /=\/=\    |  ##  ===============================================  ##  |    /=\/=\    |  ##
##  |  /========\  |  ##     License GPLv3+: GNU GPL version 3 or later    ##  |  /========\  |  ##
##  |/   ======   \|  ##          http://gnu.org/licenses/gpl.html         ##  |/   ======   \|  ##
###################################################################################################
## $Id: patd.h 4108 2014-02-06 17:10:52Z masar@masar $
#################################################################################################*/

#ifndef PATD_PATD_H
#define PATD_PATD_H 1

#include <netinet/ip.h>
#include <time.h>

//************************************************************************************************/
//* Types
//************************************************************************************************/
typedef struct _PatdIcmp
   {
   int fdIcmp;
   } PatdIcmp;

typedef struct _PatdServer
   {
   int fdServer;
   } PatdServer;

typedef struct _PatdPool
   {
   int fdPool;
   } PatdPool;

typedef struct _PatdClient
   {
   int iId;
   int fdClient;
   int iStatus;
   struct timespec tsLast;
   struct timespec tsTimeOut;
   struct sockaddr_storage saAddr;
   int iCurSeq;
   int iMaxSeq;
   int iCurRet;
   int iMaxRet;
   int iCurTTL;
   int iMaxTTL;
   int iSize;
   int iFragment;
   } PatdClient;

//************************************************************************************************/
//* Data
//************************************************************************************************/
#define MAX_CLIENTS     5 //1024
#define MAX_SEQ_BITS    6
#define MAX_SEQ_MASK    ((1 << MAX_SEQ_BITS) - 1)
#define MAX_SEQ         ((1 << MAX_SEQ_BITS) - 1)
#define MAX_TTL         ((1 << MAX_SEQ_BITS) - 1)
#define MAX_ICMP_SIZE   0x7FF0

#define COMMAND_IP4        '4'
#define COMMAND_IP4_SIZE   14
#define RESPONSE_IP4_SIZE  10

#define COMMAND_IP6        '6'
#define COMMAND_IP6_SIZE   26
#define RESPONSE_IP6_SIZE  22

#define RESULT_OK          'o'
#define RESULT_OK_LAST     'O'
#define RESULT_ERROR       'e'
#define RESULT_ERROR_LAST  'E'
#define RESULT_TRACE       't'
#define RESULT_TRACE_LAST  'T'

#define STATUS_UNUSED      0
#define STATUS_NEW         1
#define STATUS_READ_IP4    2
#define STATUS_IP4         3
#define STATUS_READ_IP6    4
#define STATUS_IP6         5

extern uint32_t   ui32_Pid;
extern uint8_t    aui8_Buffer[IP_MAXPACKET + 1];

extern PatdIcmp   patd_Icmp4;
extern PatdIcmp   patd_Icmp6;
extern PatdServer patd_Server;
extern PatdPool   patd_Pool;

/*************************************************************************************************/
/** Prototipes.
**************************************************************************************************/
void rearmClient (PatdClient * pd);
PatdClient * getPatdClient (int iId);
void deletePatdClient (PatdClient * pd);

#endif
/*************************************************************************************************/
/*************************************************************************************************/
